/* (C) 2001-2002 Andrew J. Kroll */
/* see file README for license details */
/* tcp/udp socket code */


#include <tcpvars.h>
#include <in.h>
#include <netinet.h>
#include <socket.h>
#include <ioconfig.h>
#include <tcp.h>
#include <timers.h>
#include <slip.h>
#include <tcpx.h>
#include <compat.h>

#define DEBUGD 0
#define SOCKM 0

void sendsocket(int s)
{

	register unsigned char *sktp;
	
#if DEBUGD
	printf("sendsocket %i, %i data bytes ",s, sockdlen[s]);
#endif
	sktp = SOKAT(s);
	if(sktp[STTYPE(0, 0)] == SOCK_STREAM) {
		fasttcppack(s, (sktp[BUFSTA(0, 0)] & 0xff), (sktp[SOCKST(0, 0)] & 0xff), sockdlen[s], wdatapointers[s], 0);
	}
}

void check_sockets(void)
{
	register unsigned char *sktp;
	int i;
	int mss;
	unsigned long now;

	inc32word((unsigned char *)&isnclock);
	for(i = BUFFERS - 1; i > 0; i--) {
 		now = CHK_TIME1;
		/* scheduled combined writes */
		sktp = SOKAT(i);
		if(sktp[SOCKST(0, 0)] != STATE_CLOSED) {

			if((now > timertx[i]) && (sockdlen[i] == 0) && (wbufcount[i] != 0)) {
				/* spew a waiting build packet! */
				mss = ((sktp[SOKMSS(0, 0)] * 256) + sktp[SOKMSS(0, 1)]) - 24; /* 24 is our standard header length */
				sktp[BUFSTA(0, 0)] = PSH | ACK;
				sockdlen[i] = wbufcount[i];
				if(sockdlen[i] > mss) sockdlen[i] = mss;
				errortx[i] = -1;
				sockexpecting[i] = 1;
			}
			if((now>timertx[i]) && ((sockdlen[i] != 0) || (sockexpecting[i] != 0))) {
				errortx[i]++;
				if(errortx[i] > 10) {
					/* socket timed out! ignorantly close it. */
					sktp[SOCKST(0, 0)] = STATE_CLOSED;
					sockexpecting[i] = 0;
					sockdlen[i] = 0;
				} else {
					timertx[i] = CHK_TIME1 + (loopcali * RETRYTIME);
					sendsocket(i);
				}
			}
		}
	}
}

void schedulestuff(void)
{
	check_sockets();
	TRANSPORT
}


int writetcpbytes(int fd, unsigned char *bytes, unsigned int len)
{
	register unsigned char *sktp;
	unsigned char *to;
	int status=0;
	int count;
	int mss;

	schedulestuff();

	sktp=SOKAT(fd);
	if(sktp[SOCKST(0, 0)]!=STATE_ESTABLISHED) {
		status=-1;
	} else {
		if(wbufcount[fd]==0) {
			timertx[fd]=add32word(CHK_TIME1,(BUFFERS*2));
		}
		status=len;
		if(status>(MAXTAILW-wbufcount[fd])) {
			status=MAXTAILW-wbufcount[fd];
		}
		count=wbufcount[fd];
		wbufcount[fd]+=status;
		to=wdatapointers[fd]+count;
		/* copy bytes */
		memmove(to,bytes,status);
	}
	if((sockdlen[fd] == 0) && (wbufcount[fd] != 0) && (sockexpecting[fd] == 0)) {
		mss=((sktp[SOKMSS(0, 0)]*256) + sktp[SOKMSS(0, 1)]) - 24; /* 24 is our standard header length */
		sktp[BUFSTA(0, 0)] = PSH | ACK;
		sockdlen[fd] = wbufcount[fd];
		if(sockdlen[fd] > mss) sockdlen[fd] = mss;
		sockexpecting[fd] = 1;
		timertx[fd] = CHK_TIME1 + (loopcali * RETRYTIME); 
		/* fire out this packet _NOW_ */
		sendsocket(fd);
	}
	return(status);
}

/* open, returns handle */
int sockconnect(int fd, unsigned short remoteport, unsigned char *destip)
{
	/* 
	fd points to the socket slot here. 
	I'm not going to bother storing the fd twice, 
	so it will return the fd right back upon success. 
	This simplifies alot of things.
	A higher layer should assign it a different one anyhow.
	*/
	register unsigned char *sktp;

	sktp = SOKAT(fd);
	if (sktp[SOCKST(0, 0)] != STATE_NOTINIT) {
		fd = -1;
	} else {
		rdatapointers[fd] = malloc(MAXTAILR);
		if(rdatapointers[fd] == NULL) {
#if SOCKM
			printf("socket.c: %i No RAM\n",__LINE__);
#endif
			return(-1);
		}
		wdatapointers[fd] = malloc(MAXTAILW);
		if(wdatapointers[fd] == NULL) {
			free(rdatapointers[fd]);
#if SOCKM
			printf("socket.c: %i No RAM\n", __LINE__);
#endif
			return(-1);
		}
		switch(sktp[STTYPE(0, 0)]) {
			case SOCK_STREAM: /* tcp */
				fd = tcp_open(fd, remoteport, destip);
				/* now we wait untill such a time that 
				we eiter time out, or get a connection. 
				It really needs something accurate for a timer.
				60hz would be wonderful.
				*/
				if(fd > 0) {
					do {
						schedulestuff();
					} while ((sktp[SOCKST(0, 0)] != STATE_CLOSED) && (sktp[SOCKST(0, 0)] != STATE_ESTABLISHED));
					if(sktp[SOCKST(0, 0)] != STATE_ESTABLISHED) {
						fd = -1;
					}
				}
				break;
				
			case SOCK_DGRAM: /* udp */
				schedulestuff();
				fd = -1;
				break;
			
			default:
				schedulestuff();
				fd = -1;
				break;
		}
	}
	return(fd);
}

/* write */
int sockwrite(fd, sokbuf, len)
int fd;
void *sokbuf;
unsigned int len;
{

	register unsigned char *sktp;
	unsigned char *xbuf;
	int j;
	int status=-1;

	schedulestuff();

	if((fd < 1) || (fd>BUFFERS)) {
		status=-2;
	} else {
		sktp=SOKAT(fd);
		switch(sktp[STTYPE(0, 0)]) {
			case SOCK_STREAM: /* tcp */
				if(sktp[SOCKST(0, 0)]==STATE_ESTABLISHED) {
					j=0;
					status=0;
					xbuf=(unsigned char *)sokbuf;
					while(len>0) {
						j=writetcpbytes(fd,xbuf,len);
						status+=j;
						xbuf+=j;
						len-=j;
						if(j<0) {
							len=0;
						}
					}
				}
				schedulestuff();
				break;

			case SOCK_DGRAM: /* udp */
				break;
			
			default:
				break;
		}
	}
	return(status);
}

/* read */
int sockread(fd, sokbuf, len)
int fd;
void *sokbuf;
unsigned int len;
{
	/* sneak in some tcp/ip handling here */
	register unsigned char *sktp;
	unsigned char *bleh;
	unsigned char *here;
	int i;
	int status=-1;
	int mss;

	schedulestuff();

	if((fd < 1) || (fd>BUFFERS)) {
		status=-2;
	} else {
		sktp=SOKAT(fd);
		switch(sktp[STTYPE(0, 0)]) {
			case SOCK_STREAM: /* tcp */
				if(rbufcount[fd] > 0) { /* drain socket */
					/* if there's data here, read it */
					/* for now also our sockets are non-blocking */
					bleh=sokbuf;
					here=rdatapointers[fd];
					for(status=0;((status<len) && (rbufcount[fd]!=0));status++) {
						i=rbuftail[fd];
						bleh[status]=here[i];
						rbufcount[fd]--;
						rbuftail[fd]++;
						if(rbuftail[fd]==MAXTAILR) rbuftail[fd]=0;
					}
				} else {
					if(sktp[SOCKST(0, 0)]==STATE_ESTABLISHED) {
						status=0; /* keep socket alive! *till really closed* */
					}
				}
				if(sktp[SOCKST(0, 0)] == STATE_ESTABLISHED) {
					/* extra data push if need be */
					if((sockdlen[fd] == 0) && (wbufcount[fd] > 0)) {
						mss = ((sktp[SOKMSS(0, 0)] * 256) + sktp[SOKMSS(0, 1)]) - 24; /* 24 is our standard header length */
						sktp[BUFSTA(0, 0)] = PSH | ACK;
						sockdlen[fd] = wbufcount[fd];
						if(sockdlen[fd] > mss) sockdlen[fd] = mss;
						errortx[fd] = 0;
						timertx[fd] = CHK_TIME1 + (loopcali * RETRYTIME);
						sockexpecting[fd] = 1;
						sendsocket(fd);
					} else {
						/* advertize our new depth, this should push streaming better */
						if((sockdlen[fd] == 0) && (rbufcount[fd] < (MAXTAILR / 2)) && (status > 0)) {
							sktp[BUFSTA(0, 0)] = ACK;
							errortx[fd] = 0;
							timertx[fd] = CHK_TIME1 + (loopcali * RETRYTIME);
							sendsocket(fd);
						}
					}
				}
				break;
			case SOCK_DGRAM: /* udp */
				break;
			
			default:
				break;
		}
	}
	return(status);
}

/* for fflush() */
void flushsocket(int fd)
{
	register unsigned char *sktp;
	
	if((fd < 1) || (fd>BUFFERS)) return;
	
	sktp=SOKAT(fd);
	while(((sktp[SOCKST(0, 0)] == STATE_ESTABLISHED) || (sktp[SOCKST(0, 0)] == STATE_CLOSING)) && ((wbufcount[fd] != 0) || (sockdlen[fd] != 0))) {
		schedulestuff();
	}
}

/* close/fclose WAITS TILL FULLY TIMED OUT! */
int sockclose(int fd)
{
	register unsigned char *sktp;

	if((fd < 1) || (fd>BUFFERS)) return 0;

	sktp=SOKAT(fd);
	if(sktp[SOCKST(0, 0)] < STATE_ESTABLISHED) sktp[SOCKST(0, 0)] = STATE_CLOSED;
	if(sktp[SOCKST(0, 0)] == STATE_ESTABLISHED || sktp[SOCKST(0, 0)] == STATE_CLOSING) {
		/* drain write buffer, if we can */
		flushsocket(fd);
		if(sktp[SOCKST(0, 0)] == STATE_CLOSING) {
			/* send a FIN packet with ack set, state changes here to last ack */
			sktp[SOCKST(0, 0)] = STATE_LAST_ACK;
			sockexpecting[fd] = 0;
			sktp[BUFSTA(0, 0)] = ACK | FIN;
		} else {
			/* send plain fin packet */
			sktp[BUFSTA(0, 0)] = FIN;
			sktp[SOCKST(0, 0)] = STATE_FINWAIT1;
			sockexpecting[fd] = 1;
		}
		errortx[fd] = 0;
		timertx[fd] = CHK_TIME1 + (loopcali * RETRYTIME);
		sendsocket(fd);
	}
domore:
	schedulestuff(); /* make this happen atleast ONCE */
	if((sktp[SOCKST(0, 0)] != STATE_CLOSED) && (sktp[SOCKST(0, 0)] != STATE_TIME_WAIT) && (sktp[SOCKST(0, 0)] != STATE_LAST_ACK)) goto domore;
	sktp[SOCKST(0, 0)] = STATE_CLOSED; /* we're tight on ram, close it, don't bother waiting */
	rbufcount[fd] = 0;
	free(rdatapointers[fd]);
	free(wdatapointers[fd]);
	return 0;
}

/* find a free socket area */
int findfreeslot(int type, int protocol)
{
	register unsigned char* foo;
	int i;
	int fd = -1;
	
	for(i = SOKALLOW; i > 0; i--) {
		if((sockets[SOCKST(i,0)] == STATE_CLOSED) && (rbufcount[i] == 0)) {
			fd = i;
		}
	}
	if(fd>0) {
		foo = SOKAT(fd);
		/* secure this socket NOW! */
		memset((void *)foo, 0, SOKDAT);
		foo[SOCKST(0, 0)] = STATE_NOTINIT; /* allocated, but NOT initialized! */
		foo[STTYPE(0, 0)] = type & 0xff;
		foo[STPROT(0, 0)] = protocol & 0xff;
	}
	return(fd);
}

/* setup socket */
int socket(int domain, int type, int protocol)
{
	int fd=-1;
	if(domain == AF_INET) {
		if(type == SOCK_STREAM) {
			if(protocol == 0) {
				fd = findfreeslot(type, protocol); /* TCP/IP */
			} /* protocol == 0 */
		} /* type == SOCK_STREAM */
	} /* domain == AF_INET */
	return(fd);
}


/* 
----------------------------------------------------------------------------------------------------------------------------------
Anything below this line is untested and/or incomplete stubs. 
----------------------------------------------------------------------------------------------------------------------------------
*/

/*
When writing nonportable apps, use sockconnect() for just a bit smaller code, 
this is merely a wrapper for compatability, doesn't check sanity.
*/

int connect(int sockfd, struct sockaddr_in *serv_addr, int addrlen)
{
	int retval;

	retval = sockconnect(sockfd,(unsigned short)serv_addr->sin_port, (unsigned char *) (&(serv_addr->sin_addr.s_addr)));
	if(retval > 0) retval = 0;
	return(retval);
}


/* bind creates a blank fd that we can stat via listen */
int bind(int fd, struct sockaddr *my_addr, int addrlen)
{

	register unsigned char *sktp;
	
	sktp=SOKAT(fd);	
	if(sktp[SOCKST(0, 0)] != STATE_NOTINIT) {
		return (-1);
	}
	/* 
	This is a ficticious socket state, allowing less ram via reuse. 
	don't forget to close the bind sockfd
	if your done with it to free resources! 
	*/
	sktp[SOCKST(0, 0)] = STATE_BIND;
}

/* 
	listen
	backlog is going to be a problem on a small system... 
*/
int listen(int s, int backlog)
{

}

/* accept */
int accept(int s, struct sockaddr *addr, int *addrlen)
{
	/* ... */
	schedulestuff();
	/* ... */
}

/* tell link parser who we are and where we belong

0100 libnet.lib socket.obj xxLIBRARYxx

*/

