/* (C) 2001-2002 Andrew J. Kroll */
/* see file README for license details */
/*

TCP RFC 0793
Pre-optimized for the compiler via
keeping 0x00 stores together and other tricks to help the C optimizer
I know this makes it harder to follow, but it's worth it! HONEST!

Unfortunately writes are commited instantly due to small TPA constraints.
This may change in the future.

*/
#include <tcpvars.h>
#include <in.h>
#include <netinet.h>
#include <socket.h>
#include <ioconfig.h>
#include <chcksum.h>
#include <timers.h>
#include <tcpx.h>
#include <compat.h>
#include <ip.h>

#ifdef TCP_ENABLED
#define DEBUGI 0
#define DEBUGOPTIONS 0
#define DEBUGMOVE 0
#define TCPM 0

void freebuffer(int bufnum)
{
	free(bufptrs[bufnum]);
	bufptrs[bufnum] = NULL;
	txrxflags[bufnum] = 0x00;
}


/* Create a standard IP header... */
unsigned int tcpcreateip(unsigned char *buffer, unsigned char *ptr)
{
	/* keep 0x00 stores together to help the C optimizer */
	buffer[ 1] = 0x00; /* tos */
	buffer[ 7] = 0x00;
	buffer[10] = 0x00; /* cksum */
	buffer[11] = 0x00;
	buffer[ 2] = 0x00; /* total --.  */
	buffer[ 3] = 0x14; 	 /* `-> length so far */
	buffer[ 9] = 0x06; /* tcp */
	buffer[ 6] = 0x40; /* don't fragment, please */
	buffer[ 0] = 0x45;
	buffer[ 8] = 0xff; /* ttl */
	buffer[12] = myip[0]; /* source address (me) */
	buffer[13] = myip[1];
	buffer[14] = myip[2];
	buffer[15] = myip[3];
	buffer[16] = ptr[0]; /* destination address (them) */
	buffer[17] = ptr[1];
	buffer[18] = ptr[2];
	buffer[19] = ptr[3];
	buffer[ 4] = pakid[0]; /* packet id, unique */
	buffer[ 5] = pakid[1];
	/* since we do not fragment, ever, increment it */
	bumpid();

	/* no checksumming because we add more stuff ELSEWHERE */
	return((buffer[2] * 256) + buffer[3]); /* pointer to NEXT available data area */
}

unsigned int createtcp(unsigned char *buffer, unsigned char *ptr, int datalen, unsigned char *pdata)
{
	unsigned int x;
	unsigned int y = STANDARDHEADER; /* length of standard header in octets */
	unsigned int bar = (buffer[2] * 256) + buffer[3];

	/* set mss, MRU - header(40) */
	buffer[bar + 23] = ptr[25]; /* mssLSB */
	buffer[bar + 22] = ptr[24]; /* mssMSB */
	buffer[bar + 21] = 0x04;
	buffer[bar + 20] = 0x02;
	buffer[bar + 19] = 0x00;
	buffer[bar + 18] = 0x00; /* urgent? nahhh none of our data is urgent yet :-) */
	buffer[bar + 17] = 0x00; /* checksum */
	buffer[bar + 16] = 0x00;
	buffer[bar + 15] = ptr[19]; /* window */
	buffer[bar + 14] = ptr[18]; /* window */
	buffer[bar + 13] = ptr[17]; /* flags */
	buffer[bar + 12] = (( y / 4) & 0x0f) * 16; /* data offset in words << 4 */
	/* ack # */
	buffer[bar + 11] = ptr[15];
	buffer[bar + 10] = ptr[14];
	buffer[bar +  9] = ptr[13];
	buffer[bar +  8] = ptr[12];
	/* seq # */
	buffer[bar +  7] = ptr[11];
	buffer[bar +  6] = ptr[10];
	buffer[bar +  5] = ptr[ 9];
	buffer[bar +  4] = ptr[ 8];
	/* dest port */
	buffer[bar +  3] = ptr[ 7];
	buffer[bar +  2] = ptr[ 6];
	/* src port */
	buffer[bar +  1] = ptr[ 5];
	buffer[bar     ] = ptr[ 4];


	/* attach data */
	memmove(buffer + bar + y, pdata, datalen);

	/* add to octet counts */
	x = bar + datalen+y;
	buffer[2] = (x / 256) & 0xff;
	buffer[3] = x & 0xff;

	return(datalen+y);
}


void tcpsumming(unsigned char *buffer)
{
	unsigned int sum;
	unsigned int iplen;
	unsigned int tcplen;
	unsigned int totallen;
	unsigned char fakebuf[13];

	totallen = (buffer[2] * 256) + buffer[3];
	iplen = (buffer[0] & 0x0f) * 4;
	tcplen = totallen - iplen;
	
	fakebuf[0] = buffer[12];
	fakebuf[1] = buffer[13];
	fakebuf[2] = buffer[14];
	fakebuf[3] = buffer[15];

	fakebuf[4] = buffer[16];
	fakebuf[5] = buffer[17];
	fakebuf[6] = buffer[18];
	fakebuf[7] = buffer[19];

	fakebuf[8] = 0x00;
	fakebuf[9] = 0x06;
	fakebuf[10] = (tcplen / 256 ) & 0xff;
	fakebuf[11] = tcplen & 0xff;

	/* checksum ip */
	sum = checksum(NULL, 0, buffer, iplen);
	buffer[10] = (sum / 256) & 0xff;
	buffer[11] = sum & 0xff;

	/* checksum tcp */
	sum=checksum(fakebuf, 12, &buffer[iplen], tcplen);

	buffer[iplen + 16] = (sum / 256) & 0xff;

	buffer[iplen + 17] = sum & 0xff;

}

/* build a fast reply */
int fasttcppack(int fd, int flagstype, int state, int datlen, unsigned char *datptr, unsigned int extraflags)
{
	register unsigned char *sktp;
	unsigned char *ptr;
	unsigned char *buffer;
	unsigned int hi;
	unsigned int ht;
	int i;
	int freebuf = -1;

	/* 
	If we're in the established state, let's calculate the winsize.
	Hopefully atleast the local lan remote won't frag!
	*/
	sktp = SOKAT(fd);
	if(state == STATE_ESTABLISHED) {
		/* how much buffer left? */
		sktp[WINDOW(0, 1)] = (MAXTAILR - rbufcount[fd]) & 0xff;
		sktp[WINDOW(0, 0)] = ((MAXTAILR - rbufcount[fd]) / 256) & 0xff;
	}
	/* state */
	sktp[BUFSTA(0, 0)] = flagstype & 0xff;
	sktp[SOCKST(0, 0)] = state & 0xff;
	/* build an ip packet */
	for(i = 1; i < BUFFERS; i++) {
		if((txrxflags[i] & 0xff) == 0x00) freebuf = i;
	}

	if(freebuf < 0) return(-1); /* out of buffers */

	/* malloc the buffer */
	bufptrs[freebuf] = malloc(datlen + STANDARDHEADER + 0x14);
	if(bufptrs[freebuf] == NULL) {
		/* we're screwed... */
#if TCPM
		printf("tcpx.c: %i No RAM\n",__LINE__);
#endif
		return(-1);
	}
	txrxflags[freebuf] = (ISTX | FILLBUSY); /* lock the buffer */

	ptr = SOKAT(fd);
	/* lay out ip */
	buffer = bufptrs[freebuf];
	hi=tcpcreateip(buffer, ptr);

	/* lay out tcp */
	buffer = bufptrs[freebuf];
	ht=createtcp(buffer, ptr, datlen, datptr);

	/* calculate sums */
	buffer = bufptrs[freebuf];
	tcpsumming(buffer);

	/* set the size of the entire package */
	rxtxpointer[freebuf] = hi + ht;
	/* mark socket to buffer link, since we do NOT window outbound, we can easily clear the txbuf */

	if((extraflags & SENDNOCARE) != SENDNOCARE) {
		ptr[16] = freebuf & 0xff;
	} else {
		ptr[16] = BUFFERS;
	}
	
	/* toss it in the tx que */
	txrxflags[freebuf] = (ISTX | READY | extraflags) & 0xff;
	return(fd);
}

#ifdef DOITHERE
/* rst */
int tcp_rst(int fd)
{
#if DEBUGI
	printf("RST %i ",fd);
#endif

	return(fasttcppack(fd, RST,STATE_CLOSED, 0, (unsigned char *)NULL, 0));
}

/* syn and ack */
int tcp_synandack(int fd)
{
#if DEBUGI
	printf("SYNACK %i ",fd);
#endif

	return(fasttcppack(fd, SYN | ACK ,STATE_SYN_RECEIVED ,0 ,(unsigned char *)NULL, 0));
}

/* ack syn */
int tcp_acksyn(int fd)
{
#if DEBUGI
	printf("ACKSYN %i ",fd);
#endif

	return(fasttcppack(fd, ACK, STATE_ESTABLISHED, 0, (unsigned char *)NULL, SENDNOCARE));
}

/* Build a SYN packet */
int tcp_syn(int fd)
{
#if DEBUGI
		printf("SYN %i ",fd);
#endif

	return(fasttcppack(fd, SYN,STATE_SYN_SENT, 0, (unsigned char *)NULL, 0));
}

int tcp_psh(int fd, int len, unsigned char *ptr)
{
#if DEBUGI
		printf("PSH %i %i bytes", fd, len);
#endif

	return(fasttcppack(fd, PSH, STATE_ESTABLISHED, len, ptr, SENDNOCARE));
}

#endif DOITHERE

/* open */
int tcp_open(int fd, unsigned int remote, unsigned char *ip)
{
	register unsigned char *sktp;
	unsigned char *ptr;
	int i;
	printf("Remote: %i\n", remote);

	ptr = SOKAT(fd);
	/* memset(ptr, 0, SOKDAT); */
	/* lay in the data */
	/* dest ip */
	ptr[0] = ip[0];
	ptr[1] = ip[1];
	ptr[2] = ip[2];
	ptr[3] = ip[3];
	newport:

	/* src port start at 32768 and work up */
	/* is current port available? */
	for(i = 0; i < SOKALLOW; i++) {
		sktp = SOKAT(i);
		if( (sktp[SRCPRT(0, 0)] == ((portcounter / 256) & 0xff)) && (sktp[SRCPRT(0, 1)] == (portcounter & 0xff))) {
			portcounter++;
			if(portcounter > 65000) portcounter = 32768;
			goto newport;
		}

	}

	ptr[4] = (portcounter / 256) & 0xff;
	ptr[5] = portcounter & 0xff;
	portcounter++;

	if(portcounter > 65000) portcounter = 32768;
	/* dest port */
	ptr[6] = (remote >> 8) & 0xff;
	ptr[7] = remote & 0xff;
	/* initial seq/ack */
	sav32word(&ptr[8], isnclock);
	sav32word(&ptr[12], isnclock);

	/* window 1 buffersize */

/* lets nuke local frags! */
	ptr[18] = ((MAXPKTSZ) / 256) & 0xff;;
	ptr[19] = (MAXPKTSZ) & 0xff;

	/* fd */
	ptr[20] = fd & 0xff;

	ptr[24] = ((MAXMTUMRU) / 256) & 0xff; /* mssMSB */
	ptr[25] = (MAXMTUMRU) & 0xff; /* mssLSB */

	tcp_syn(fd);
	/* sock state */
	timertx[fd] = CHK_TIME1 + (loopcali * RETRYTIME);
	sockexpecting[fd] = 1;
	return (fd);
}


int matchsocket(unsigned char *port)
{
	register unsigned char *sktp;
	int i;
	int match = -1;

	for(i = SOKALLOW; i > 0; i--) {
		sktp = SOKAT(i);
		if(sktp[SOCKST(0, 0)] > STATE_NOTINIT) {
			/* is socket active... inline comparason */
		  if(
		     (port[0] == sktp[DESPRT(0, 0)]) && 
		     (port[1] == sktp[DESPRT(0, 1)]) && 
		     (port[2] == sktp[SRCPRT(0, 0)]) && 
		     (port[3] == sktp[SRCPRT(0, 1)])) match = i;
		  /*			printf("Socket: %i Match: %i\n"
			       "%1.1x=%1.1x %1.1x=%1.1x\n"
			       "%1.1x=%1.1x %1.1x=%1.1x\n", i,  match, 
					
			       port[0], sktp[DESPRT(0,0)], 
			       port[1], sktp[DESPRT(0,1)], 
			       port[2], sktp[SRCPRT(0,0)], 
			       port[3], sktp[SRCPRT(0,1)]); */
		}
	}
	if(match == -1) {
		for(i = SOKALLOW; i > 0; i--) {
			sktp = SOKAT(i);
			if(sktp[SOCKST(0, 0)] > STATE_NOTINIT) {
			/* is socket passive... inline comparason */
				if ((port[0] == sktp[DESPRT(0, 0)]) && (port[1] == sktp[DESPRT(0, 1)]) && (sktp[SOCKST(0, 0)] == STATE_LISTEN)) match = i;
			}
		}
	}
	return(match);
}


/* check ack match seq */
int checkack(unsigned char *ours, unsigned char *thiers)
{
	if((((ours[0] == thiers[0]) && (ours[1] == thiers[1])) && ((ours[2] == thiers[2]) && (ours[3] == thiers[3])))) return(1);
	return(0);
}


void snag32(unsigned char *from, unsigned char *to)
{
	/* copy block of 4, incase bcopy dosen't exist... */
	to[0] = from[0];
	to[1] = from[1];
	to[2] = from[2];
	to[3] = from[3];
}

void snag16(unsigned char *from, unsigned char *to)
{
	/* copy block of 2, incase bcopy dosen't exist... */
	to[0] = from[0];
	to[1] = from[1];
}


void checkoptions(unsigned char *buffer, int sock, int iplen, int tcplen)
{
	register unsigned char *sktp;
	unsigned char *thiers;
	int width;

#if DEBUGOPTIONS
	int i;

	printf("OPTIONS { ");
#endif
	sktp = SOKAT(sock);
	/* if the remote mss is less, copy it */
	thiers = buffer + iplen + 20; /* options block, pick out the mss option */
	for(width = 0; width < tcplen - 20;) {
		if(thiers[width] == 0x00) { /* end options */
#if DEBUGOPTIONS
			printf("END ");
#endif
			width = tcplen;
		} else if(thiers[width] == 0x01) { /* NOP */
#if DEBUGOPTIONS
			printf("NOP ");
#endif
			width++;
		} else if(thiers[width] == 0x02) { /* MSS */
#if DEBUGOPTIONS
			printf("MSS %i ",(thiers[width + 2] * 256) + thiers[width + 3]);
#endif
			if(((sktp[SOKMSS(0, 0)] * 256) + sktp[SOKMSS(0, 1)]) > ((thiers[width + 2] * 256) + thiers[width + 3])) {
				/* thier mss is smaller! */
				sktp[SOKMSS(0, 0)] = thiers[width + 2];
				sktp[SOKMSS(0, 1)] = thiers[width + 3];
			}
			width += thiers[width + 1];
		} else { /* unknown option... emmm lessee.... */
#if DEBUGOPTIONS
			printf("OPTION ???: [ ");
			for(i = 0; i < (thiers[width + 1] & 0xff); i++) printf("%2.2x ",thiers[width + i]);
			printf("]");
#endif
			width += thiers[width + 1];
		}
	}
#if DEBUGOPTIONS
	printf("} ");
#endif
}


void bufferdatamove(unsigned char *databuffer, unsigned int sock, unsigned int datalen)
{
	unsigned int width;
	unsigned char *bleh;
	unsigned char *there;
	bleh = databuffer;
	there = rdatapointers[sock];
#if DEBUGMOVE
	printf("INBOUND DATA Socket %i, %i Bytes [ ", sock, datalen);
#endif
	for(width = 0; width < datalen; width++) {
		there[rbufhead[sock]] = *bleh++;
#if DEBUGMOVE
		printf("%2.2x ",databuffer[width] & 0xff);
#endif
		rbufcount[sock]++;
		rbufhead[sock]++;
		if(rbufhead[sock] == MAXTAILR) rbufhead[sock] = 0;
	}
#if DEBUGMOVE
	printf("]\n");
#endif
}

/*

destination ip		[4] 0 1 2 3
source port		[2] 4 5 
destination port	[2] 6 7
sequence number		[4] 8 9 10 11
acknowledgement number	[4] 12 13 14 15
buffer used		[1] 16 
state			[1] 17
window			[2] 18 19
fd ( ==socket for demo)	[1] 20
socket state		[1] 21
type			[1] 22
protocol		[1] 23
mss			[2] 24 25

CTMP5.OUT:1086: Arithmetic overflow
CTMP5.OUT:1125: Arithmetic overflow
*/

#endif

/* tell link parser who we are and where we belong

0300 libnet.lib tcpx.obj xxLIBRARYxx

*/

